import numpy as np

import torch
import torch.nn as nn


class ValueNorm(nn.Module):
    """ Normalize a vector of observations - across the first norm_axes dimensions"""

    def __init__(self, input_shape, norm_axes=1, beta=0.99999, per_element_update=False, epsilon=1e-5,
                 device=torch.device("cpu")):
        super(ValueNorm, self).__init__()

        self.input_shape = input_shape
        self.norm_axes = norm_axes
        self.epsilon = epsilon
        self.beta = beta
        self.per_element_update = per_element_update
        self.tpdv = dict(dtype=torch.float32, device=device)

        self.running_mean = nn.Parameter(torch.zeros(input_shape), requires_grad=False).to(**self.tpdv)
        self.running_mean_sq = nn.Parameter(torch.zeros(input_shape), requires_grad=False).to(**self.tpdv)
        self.debiasing_term = nn.Parameter(torch.tensor(0.0), requires_grad=False).to(**self.tpdv)

        self.reset_parameters()

    def reset_parameters(self):
        self.running_mean.zero_()
        self.running_mean_sq.zero_()
        self.debiasing_term.zero_()

    def running_mean_var(self):
        debiased_mean = self.running_mean / self.debiasing_term.clamp(min=self.epsilon)
        debiased_mean_sq = self.running_mean_sq / self.debiasing_term.clamp(min=self.epsilon)
        debiased_var = (debiased_mean_sq - debiased_mean ** 2).clamp(min=1e-2)
        return debiased_mean, debiased_var

    @torch.no_grad()
    def update(self, input_vector):
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)

        batch_mean = input_vector.mean(dim=tuple(range(self.norm_axes)))
        batch_sq_mean = (input_vector ** 2).mean(dim=tuple(range(self.norm_axes)))

        if self.per_element_update:
            batch_size = np.prod(input_vector.size()[:self.norm_axes])
            weight = self.beta ** batch_size
        else:
            weight = self.beta

        self.running_mean.mul_(weight).add_(batch_mean * (1.0 - weight))
        self.running_mean_sq.mul_(weight).add_(batch_sq_mean * (1.0 - weight))
        self.debiasing_term.mul_(weight).add_(1.0 * (1.0 - weight))

    def normalize(self, input_vector):
        # Make sure input is float32
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)

        mean, var = self.running_mean_var()
        out = (input_vector - mean[(None,) * self.norm_axes]) / torch.sqrt(var)[(None,) * self.norm_axes]

        return out

    def denormalize(self, input_vector):
        """ Transform normalized data back into original distribution """
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)

        mean, var = self.running_mean_var()
        out = input_vector * torch.sqrt(var)[(None,) * self.norm_axes] + mean[(None,) * self.norm_axes]

        out = out.cpu().numpy()

        return out


import torch.nn as nn
from .util import init, get_clones

"""MLP modules."""

class MLPLayer(nn.Module):
    def __init__(self, input_dim, hidden_size, layer_N, use_orthogonal, use_ReLU):
        super(MLPLayer, self).__init__()
        self._layer_N = layer_N

        active_func = [nn.Tanh(), nn.ReLU()][use_ReLU]
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        gain = nn.init.calculate_gain(['tanh', 'relu'][use_ReLU])

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain=gain)

        self.fc1 = nn.Sequential(
            init_(nn.Linear(input_dim, hidden_size)), active_func, nn.LayerNorm(hidden_size))
        self.fc_h = nn.Sequential(init_(
            nn.Linear(hidden_size, hidden_size)), active_func, nn.LayerNorm(hidden_size))
        self.fc2 = get_clones(self.fc_h, self._layer_N)

    def forward(self, x):
        x = self.fc1(x)
        for i in range(self._layer_N):
            x = self.fc2[i](x)
        return x


class MLPBase(nn.Module):
    def __init__(self, args, obs_shape, cat_self=True, attn_internal=False):
        super(MLPBase, self).__init__()

        self._use_feature_normalization = args.use_feature_normalization
        self._use_orthogonal = args.use_orthogonal
        self._use_ReLU = args.use_ReLU
        self._stacked_frames = args.stacked_frames
        self._layer_N = args.layer_N
        self.hidden_size = args.mlp_hidden_dim

        # obs_dim = obs_shape[0]
        obs_dim = obs_shape
        if self._use_feature_normalization:
            self.feature_norm = nn.LayerNorm(obs_dim)

        self.mlp = MLPLayer(obs_dim, self.hidden_size,
                              self._layer_N, self._use_orthogonal, self._use_ReLU)

    def forward(self, x):
        if self._use_feature_normalization:
            x = self.feature_norm(x)

        x = self.mlp(x)

        return x


from .distributions import Bernoulli, Categorical, DiagGaussian
import torch
import torch.nn as nn

class ACTLayer(nn.Module):
    """
    MLP Module to compute actions.
    :param action_space: (gym.Space) action space.
    :param inputs_dim: (int) dimension of network input.
    :param use_orthogonal: (bool) whether to use orthogonal initialization.
    :param gain: (float) gain of the output layer of the network.
    """
    def __init__(self, n_actions, inputs_dim, use_orthogonal, gain):
        super(ACTLayer, self).__init__()
        self.mixed_action = False
        self.multi_discrete = False
        self.action_out = Categorical(inputs_dim, n_actions, use_orthogonal, gain)


        # if action_space.__class__.__name__ == "Discrete":
        #     action_dim = action_space.n
        #     self.action_out = Categorical(inputs_dim, action_dim, use_orthogonal, gain)
        # elif action_space.__class__.__name__ == "Box":
        #     action_dim = action_space.shape[0]
        #     self.action_out = DiagGaussian(inputs_dim, action_dim, use_orthogonal, gain)
        # elif action_space.__class__.__name__ == "MultiBinary":
        #     action_dim = action_space.shape[0]
        #     self.action_out = Bernoulli(inputs_dim, action_dim, use_orthogonal, gain)
        # elif action_space.__class__.__name__ == "MultiDiscrete":
        #     self.multi_discrete = True
        #     action_dims = action_space.high - action_space.low + 1
        #     self.action_outs = []
        #     for action_dim in action_dims:
        #         self.action_outs.append(Categorical(inputs_dim, action_dim, use_orthogonal, gain))
        #     self.action_outs = nn.ModuleList(self.action_outs)
        # else:  # discrete + continous
        #     self.mixed_action = True
        #     continous_dim = action_space[0].shape[0]
        #     discrete_dim = action_space[1].n
        #     self.action_outs = nn.ModuleList([DiagGaussian(inputs_dim, continous_dim, use_orthogonal, gain), Categorical(
        #         inputs_dim, discrete_dim, use_orthogonal, gain)])
    
    def forward(self, x, available_actions=None, deterministic=False):
        """
        Compute actions and action logprobs from given input.
        :param x: (torch.Tensor) input to network.
        :param available_actions: (torch.Tensor) denotes which actions are available to agent
                                  (if None, all actions available)
        :param deterministic: (bool) whether to sample from action distribution or return the mode.

        :return actions: (torch.Tensor) actions to take.
        :return action_log_probs: (torch.Tensor) log probabilities of taken actions.
        """
        # if self.mixed_action :
        #     actions = []
        #     action_log_probs = []
        #     for action_out in self.action_outs:
        #         action_logit = action_out(x)
        #         action = action_logit.mode() if deterministic else action_logit.sample()
        #         action_log_prob = action_logit.log_probs(action)
        #         actions.append(action.float())
        #         action_log_probs.append(action_log_prob)

        #     actions = torch.cat(actions, -1)
        #     action_log_probs = torch.sum(torch.cat(action_log_probs, -1), -1, keepdim=True)

        # elif self.multi_discrete:
        #     actions = []
        #     action_log_probs = []
        #     for action_out in self.action_outs:
        #         action_logit = action_out(x)
        #         action = action_logit.mode() if deterministic else action_logit.sample()
        #         action_log_prob = action_logit.log_probs(action)
        #         actions.append(action)
        #         action_log_probs.append(action_log_prob)

        #     actions = torch.cat(actions, -1)
        #     action_log_probs = torch.cat(action_log_probs, -1)
        
        # else:
        action_logits = self.action_out(x, available_actions)
        actions = action_logits.mode() if deterministic else action_logits.sample() 
        action_log_probs = action_logits.log_probs(actions)
        
        return actions, action_log_probs

    def get_probs(self, x, available_actions=None):
        """
        Compute action probabilities from inputs.
        :param x: (torch.Tensor) input to network.
        :param available_actions: (torch.Tensor) denotes which actions are available to agent
                                  (if None, all actions available)

        :return action_probs: (torch.Tensor)
        """
        # if self.mixed_action or self.multi_discrete:
        #     action_probs = []
        #     for action_out in self.action_outs:
        #         action_logit = action_out(x)
        #         action_prob = action_logit.probs
        #         action_probs.append(action_prob)
        #     action_probs = torch.cat(action_probs, -1)
        # else:
        action_logits = self.action_out(x, available_actions)
        action_probs = action_logits.probs
        
        return action_probs

    def evaluate_actions(self, x, action, available_actions=None):
        """
        Compute log probability and entropy of given actions.
        :param x: (torch.Tensor) input to network.
        :param action: (torch.Tensor) actions whose entropy and log probability to evaluate.
        :param available_actions: (torch.Tensor) denotes which actions are available to agent
                                                              (if None, all actions available)
        :param active_masks: (torch.Tensor) denotes whether an agent is active or dead.

        :return action_log_probs: (torch.Tensor) log probabilities of the input actions.
        :return dist_entropy: (torch.Tensor) action distribution entropy for the given inputs.
        """
        # if self.mixed_action:
        #     a, b = action.split((2, 1), -1)
        #     b = b.long()
        #     action = [a, b] 
        #     action_log_probs = [] 
        #     dist_entropy = []
        #     for action_out, act in zip(self.action_outs, action):
        #         action_logit = action_out(x)
        #         action_log_probs.append(action_logit.log_probs(act))
        #         if active_masks is not None:
        #             if len(action_logit.entropy().shape) == len(active_masks.shape):
        #                 dist_entropy.append((action_logit.entropy() * active_masks).sum()/active_masks.sum()) 
        #             else:
        #                 dist_entropy.append((action_logit.entropy() * active_masks.squeeze(-1)).sum()/active_masks.sum())
        #         else:
        #             dist_entropy.append(action_logit.entropy().mean())
                
        #     action_log_probs = torch.sum(torch.cat(action_log_probs, -1), -1, keepdim=True)
        #     dist_entropy = dist_entropy[0] / 2.0 + dist_entropy[1] / 0.98 #! dosen't make sense

        # elif self.multi_discrete:
        #     action = torch.transpose(action, 0, 1)
        #     action_log_probs = []
        #     dist_entropy = []
        #     for action_out, act in zip(self.action_outs, action):
        #         action_logit = action_out(x)
        #         action_log_probs.append(action_logit.log_probs(act))
        #         if active_masks is not None:
        #             dist_entropy.append((action_logit.entropy()*active_masks.squeeze(-1)).sum()/active_masks.sum())
        #         else:
        #             dist_entropy.append(action_logit.entropy().mean())

        #     action_log_probs = torch.cat(action_log_probs, -1) # ! could be wrong
        #     dist_entropy = torch.tensor(dist_entropy).mean()
        
        # else:

        # if agent is dead, only available action is no-op, making active masks unnecessary
        action_logits = self.action_out(x, available_actions)  
        action_log_probs = action_logits.log_probs(action)

        # if active_masks is not None:
        #     dist_entropy = (action_logits.entropy()*active_masks.squeeze(-1)).sum()/active_masks.sum()
        dist_entropy = action_logits.entropy().mean()
        
        return action_log_probs, dist_entropy


import torch as th
import torch.nn.functional as F
from torch.distributions import Categorical
from .epsilon_schedules import DecayThenFlatSchedule, FlatSchedule


REGISTRY = {}


class MultinomialActionSelector():
    def __init__(self, args):
        self.args = args
        self.test_greedy = getattr(args, "test_greedy", True)

    def select_action(self, agent_inputs, avail_actions, t_env, t_ep, test_mode=False):
        q_values = agent_inputs.clone()

        if self.args.algo_name == "isql":
            values = self.args.alpha * th.logsumexp(q_values / self.args.alpha, dim=2, keepdims=True)
            logits = 1. / self.args.alpha * (q_values - values)
            logits[avail_actions == 0.0] = -1e10
            exp_logits = th.exp(logits) # shape: (1, n_agents, n_actions)

            picked_actions = Categorical(exp_logits).sample().long()

        else:        
            q_values[avail_actions == 0.0] = 0.0
            if test_mode and self.test_greedy:
                picked_actions = q_values.max(dim=2)[1]
            else:
                picked_actions = Categorical(q_values).sample().long()

        return picked_actions


REGISTRY["multinomial"] = MultinomialActionSelector


class EpsilonGreedyActionSelector():

    def __init__(self, args):
        self.args = args

        self.schedule = DecayThenFlatSchedule(args.epsilon_start, args.epsilon_finish, args.epsilon_anneal_time,
                                              decay="linear")
        self.epsilon = self.schedule.eval(0)

    def select_action(self, agent_inputs, avail_actions, t_env, test_mode=False):

        # Assuming agent_inputs is a batch of Q-Values for each agent bav
        self.epsilon = self.schedule.eval(t_env)

        if test_mode:
            # Greedy action selection only
            self.epsilon = 0.0
        # mask actions that are excluded from selection
        masked_q_values = agent_inputs.clone()
        masked_q_values[avail_actions == 0.0] = -float("inf")  # should never be selected!

        random_numbers = th.rand_like(agent_inputs[:, :, 0])
        pick_random = (random_numbers < self.epsilon).long()
        random_actions = Categorical(avail_actions.float()).sample().long()

        picked_actions = pick_random * random_actions + (1 - pick_random) * masked_q_values.max(dim=2)[1]
        return picked_actions


REGISTRY["epsilon_greedy"] = EpsilonGreedyActionSelector


import sys
import os
from os.path import dirname, abspath
import collections
from copy import deepcopy
import yaml

import numpy as np
from sacred import Experiment, SETTINGS
from sacred.observers import FileStorageObserver
from sacred.utils import apply_backspaces_and_linefeeds
import torch as th

from utils.logging import get_logger
from run_ippo import run as run_ippo

SETTINGS['CONFIG']['READ_ONLY_CONFIG'] = False
SETTINGS['CAPTURE_MODE'] = 'sys'
os.environ['KMP_DUPLICATE_LIB_OK'] = 'True'
logger = get_logger()

ex = Experiment("Highway_iPLAN")
ex.logger = logger
ex.captured_out_filter = apply_backspaces_and_linefeeds

results_path = "results"

@ex.main
def my_main(_run, _config, _log):
    # Setting the random seed throughout the modules
    config = config_copy(_config)

    np.random.seed(config["seed"])
    th.manual_seed(config["seed"])
    config['env_args']['seed'] = config["seed"] # overrides env_args seed
    # run the framework
    run_ippo(_run, config, _log)


def _get_config(config_name, subfolder):
    if config_name is not None:
        with open(os.path.join(os.path.dirname(__file__), "config", subfolder, "{}.yaml".format(config_name)), "r") as f:
            try:
                config_dict = yaml.safe_load(f)
            except yaml.YAMLError as exc:
                assert False, "{}.yaml error: {}".format(config_name, exc)
        return config_dict

def set_arg(config_dict, params, arg_name, arg_type):
    for _i, _v in enumerate(params):
        if _v.split("=")[0] == arg_name:
            arg_name = _v.split("=")[0].replace("--", "")
            arg_value = _v.split("=")[1]
            config_dict[arg_name] = arg_type(arg_value)
            del params[_i]
            return config_dict


def recursive_dict_update(d, u):
    '''update dict d with items in u recursively. if key present in both d and u, 
    value in d takes precedence. 
    '''
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            d[k] = recursive_dict_update(d.get(k, {}), v)
        else:
            if k not in d.keys():
                d[k] = v                
    return d


def config_copy(config):
    if isinstance(config, dict):
        return {k: config_copy(v) for k, v in config.items()}
    elif isinstance(config, list):
        return [config_copy(v) for v in config]
    else:
        return deepcopy(config)


if __name__ == '__main__':
    # Get the defaults from default.yaml
    config_dict = _get_config("default",  "")

    # Load env base configs when considering difficulties
    if config_dict["env"] == "highway":
        env_config = _get_config("highway", "envs")
    elif config_dict["env"] == "MPE":
        if config_dict["difficulty"] == "easy":
            env_config = _get_config("simple_spread_Hetero", "envs")
        else:
            env_config = _get_config("simple_spread_Hetero_H", "envs")

    # Load algorithm configs
    alg_config = _get_config("ippo", "algs")

    # update env_config and alg_config with values in config dict 
    # copy modified env args for logging purpose 
    config_dict = recursive_dict_update(config_dict, env_config)
    config_dict = recursive_dict_update(config_dict, alg_config)

    # add config to sacred
    ex.add_config(config_dict)

    # Save to disk by default for sacred
    logger.info("Saving to FileStorageObserver in results/sacred.")
    file_obs_path = os.path.join(results_path, "sacred")
    ex.observers.append(FileStorageObserver.create(file_obs_path))

    ex.run_commandline()




import torch as th
import numpy as np
from types import SimpleNamespace as SN


class EpisodeBatch:
    def __init__(self,
                 scheme,
                 groups,
                 batch_size,
                 max_seq_length,
                 data=None,
                 preprocess=None,
                 device="cpu"):
        self.scheme = scheme.copy()
        self.groups = groups
        self.batch_size = batch_size
        self.max_seq_length = max_seq_length
        self.preprocess = {} if preprocess is None else preprocess
        self.device = device

        if data is not None:
            self.data = data
        else:
            self.data = SN()
            self.data.transition_data = {}
            self.data.episode_data = {}
            self._setup_data(self.scheme, self.groups, batch_size, max_seq_length, self.preprocess)

    def _setup_data(self, scheme, groups, batch_size, max_seq_length, preprocess):
        if preprocess is not None:
            for k in preprocess:
                assert k in scheme
                new_k = preprocess[k][0]
                transforms = preprocess[k][1]

                vshape = self.scheme[k]["vshape"]
                dtype = self.scheme[k]["dtype"]
                for transform in transforms:
                    vshape, dtype = transform.infer_output_info(vshape, dtype)

                self.scheme[new_k] = {
                    "vshape": vshape,
                    "dtype": dtype
                }
                if "group" in self.scheme[k]:
                    self.scheme[new_k]["group"] = self.scheme[k]["group"]
                if "episode_const" in self.scheme[k]:
                    self.scheme[new_k]["episode_const"] = self.scheme[k]["episode_const"]

        assert "filled" not in scheme, '"filled" is a reserved key for masking.'
        scheme.update({
            "filled": {"vshape": (1,), "dtype": th.long},
        })

        for field_key, field_info in scheme.items():
            assert "vshape" in field_info, "Scheme must define vshape for {}".format(field_key)
            vshape = field_info["vshape"]
            episode_const = field_info.get("episode_const", False)
            group = field_info.get("group", None)
            dtype = field_info.get("dtype", th.float32)

            if isinstance(vshape, int):
                vshape = (vshape,)

            if group:
                assert group in groups, "Group {} must have its number of members defined in _groups_".format(group)
                shape = (groups[group], *vshape)
            else:
                shape = vshape

            if episode_const:
                self.data.episode_data[field_key] = th.zeros((batch_size, *shape), dtype=dtype, device=self.device)
            else:
                self.data.transition_data[field_key] = th.zeros((batch_size, max_seq_length, *shape), dtype=dtype, device=self.device)

    def extend(self, scheme, groups=None):
        self._setup_data(scheme, self.groups if groups is None else groups, self.batch_size, self.max_seq_length)

    def to(self, device):
        for k, v in self.data.transition_data.items():
            self.data.transition_data[k] = v.to(device)
        for k, v in self.data.episode_data.items():
            self.data.episode_data[k] = v.to(device)
        self.device = device

    def update(self, data, bs=slice(None), ts=slice(None), mark_filled=True):
        slices = self._parse_slices((bs, ts))
        for k, v in data.items():
            if k in self.data.transition_data:
                target = self.data.transition_data
                if mark_filled:
                    target["filled"][slices] = 1
                    mark_filled = False
                _slices = slices
            elif k in self.data.episode_data:
                target = self.data.episode_data
                _slices = slices[0]
            else:
                raise KeyError("{} not found in transition or episode data".format(k))

            dtype = self.scheme[k].get("dtype", th.float32)
            v = th.tensor(v, dtype=dtype, device=self.device)
            self._check_safe_view(v, target[k][_slices])
            target[k][_slices] = v.view_as(target[k][_slices])

            if k in self.preprocess:
                new_k = self.preprocess[k][0]
                v = target[k][_slices]
                for transform in self.preprocess[k][1]:
                    v = transform.transform(v)
                target[new_k][_slices] = v.view_as(target[new_k][_slices])

    def _check_safe_view(self, v, dest):
        idx = len(v.shape) - 1
        for s in dest.shape[::-1]:
            if v.shape[idx] != s:
                if s != 1:
                    raise ValueError("Unsafe reshape of {} to {}".format(v.shape, dest.shape))
            else:
                idx -= 1

    def __getitem__(self, item):
        if isinstance(item, str):
            if item in self.data.episode_data:
                return self.data.episode_data[item]
            elif item in self.data.transition_data:
                return self.data.transition_data[item]
            else:
                raise ValueError
        elif isinstance(item, tuple) and all([isinstance(it, str) for it in item]):
            new_data = self._new_data_sn()
            for key in item:
                if key in self.data.transition_data:
                    new_data.transition_data[key] = self.data.transition_data[key]
                elif key in self.data.episode_data:
                    new_data.episode_data[key] = self.data.episode_data[key]
                else:
                    raise KeyError("Unrecognised key {}".format(key))

            # Update the scheme to only have the requested keys
            new_scheme = {key: self.scheme[key] for key in item}
            new_groups = {self.scheme[key]["group"]: self.groups[self.scheme[key]["group"]]
                          for key in item if "group" in self.scheme[key]}
            ret = EpisodeBatch(new_scheme, new_groups, self.batch_size, self.max_seq_length, data=new_data, device=self.device)
            return ret
        else:
            item = self._parse_slices(item)
            new_data = self._new_data_sn()
            for k, v in self.data.transition_data.items():
                new_data.transition_data[k] = v[item]
            for k, v in self.data.episode_data.items():
                new_data.episode_data[k] = v[item[0]]

            ret_bs = self._get_num_items(item[0], self.batch_size)
            ret_max_t = self._get_num_items(item[1], self.max_seq_length)

            ret = EpisodeBatch(self.scheme, self.groups, ret_bs, ret_max_t, data=new_data, device=self.device)
            return ret

    def _get_num_items(self, indexing_item, max_size):
        if isinstance(indexing_item, list) or isinstance(indexing_item, np.ndarray):
            return len(indexing_item)
        elif isinstance(indexing_item, slice):
            _range = indexing_item.indices(max_size)
            return 1 + (_range[1] - _range[0] - 1)//_range[2]

    def _new_data_sn(self):
        new_data = SN()
        new_data.transition_data = {}
        new_data.episode_data = {}
        return new_data

    def _parse_slices(self, items):
        parsed = []
        # Only batch slice given, add full time slice
        if (isinstance(items, slice)  # slice a:b
            or isinstance(items, int)  # int i
            or (isinstance(items, (list, np.ndarray, th.LongTensor, th.cuda.LongTensor)))  # [a,b,c]
            ):
            items = (items, slice(None))

        # Need the time indexing to be contiguous
        if isinstance(items[1], list):
            raise IndexError("Indexing across Time must be contiguous")

        for item in items:
            #TODO: stronger checks to ensure only supported options get through
            if isinstance(item, int):
                # Convert single indices to slices
                parsed.append(slice(item, item+1))
            else:
                # Leave slices and lists as is
                parsed.append(item)
        return parsed

    def max_t_filled(self):
        return th.sum(self.data.transition_data["filled"], 1).max(0)[0]

    def __repr__(self):
        return "EpisodeBatch. Batch Size:{} Max_seq_len:{} Keys:{} Groups:{}".format(self.batch_size,
                                                                                     self.max_seq_length,
                                                                                     self.scheme.keys(),
                                                                                     self.groups.keys())


class ReplayBuffer(EpisodeBatch):
    def __init__(self, scheme, groups, buffer_size, max_seq_length, preprocess=None, device="cpu"):
        super(ReplayBuffer, self).__init__(scheme, groups, buffer_size, max_seq_length, preprocess=preprocess, device=device)
        self.buffer_size = buffer_size  # same as self.batch_size but more explicit
        self.buffer_index = 0
        self.episodes_in_buffer = 0

    def insert_episode_batch(self, ep_batch):
        if self.buffer_index + ep_batch.batch_size <= self.buffer_size:
            self.update(ep_batch.data.transition_data,
                        slice(self.buffer_index, self.buffer_index + ep_batch.batch_size),
                        slice(0, ep_batch.max_seq_length),
                        mark_filled=False)
            self.update(ep_batch.data.episode_data,
                        slice(self.buffer_index, self.buffer_index + ep_batch.batch_size))
            self.buffer_index = (self.buffer_index + ep_batch.batch_size)
            self.episodes_in_buffer = max(self.episodes_in_buffer, self.buffer_index)
            self.buffer_index = self.buffer_index % self.buffer_size
            assert self.buffer_index < self.buffer_size
        else:
            buffer_left = self.buffer_size - self.buffer_index
            self.insert_episode_batch(ep_batch[0:buffer_left, :])
            self.insert_episode_batch(ep_batch[buffer_left:, :])

    def can_sample(self, batch_size):
        return self.episodes_in_buffer >= batch_size

    def sample(self, batch_size):
        assert self.can_sample(batch_size)
        if self.episodes_in_buffer == batch_size:
            return self[:batch_size]
        else:
            # Uniform sampling only atm
            ep_ids = np.random.choice(self.episodes_in_buffer, batch_size, replace=False)
            return self[ep_ids]

    def __repr__(self):
        return "ReplayBuffer. {}/{} episodes. Keys:{} Groups:{}".format(self.episodes_in_buffer,
                                                                        self.buffer_size,
                                                                        self.scheme.keys(),
                                                                        self.groups.keys())




from collections import defaultdict
import logging
import numpy as np

class Logger:
    def __init__(self, console_logger):
        self.console_logger = console_logger

        self.use_tb = False
        self.use_sacred = False
        self.use_hdf = False

        self.stats = defaultdict(lambda: [])

    def setup_tb(self, directory_name):
        # Import here so it doesn't have to be installed if you don't use it
        from tensorboard_logger import configure, log_value
        configure(directory_name)
        self.tb_logger = log_value
        self.use_tb = True

    def setup_sacred(self, sacred_run_dict):
        self.sacred_info = sacred_run_dict.info
        self.use_sacred = True

    def log_stat(self, key, value, t, to_sacred=True):
        self.stats[key].append((t, value))

        if self.use_tb:
            self.tb_logger(key, value, t)

        if self.use_sacred and to_sacred:
            if key in self.sacred_info:
                self.sacred_info["{}_T".format(key)].append(t)
                self.sacred_info[key].append(value)
            else:
                self.sacred_info["{}_T".format(key)] = [t]
                self.sacred_info[key] = [value]

    def print_recent_stats(self):
        import torch as th
        log_str = "Recent Stats | t_env: {:>10} | Episode: {:>8}\n".format(*self.stats["episode"][-1])
        i = 0
        for (k, v) in sorted(self.stats.items()):
            if k == "episode":
                continue
            i += 1
            window = 5 if k != "epsilon" else 1
            item = "{:.4f}".format(th.mean(th.tensor([x[1] for x in self.stats[k][-window:]])))
            log_str += "{:<25}{:>8}".format(k + ":", item)
            log_str += "\n" if i % 4 == 0 else "\t"
        self.console_logger.info(log_str)


# set up a custom logger
def get_logger():
    logger = logging.getLogger()
    logger.handlers = []
    ch = logging.StreamHandler()
    formatter = logging.Formatter('[%(levelname)s %(asctime)s] %(name)s %(message)s', '%H:%M:%S')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.setLevel('DEBUG')

    return logger




import numpy as np
from collections import deque

# Multi-threading version
# Highway obs, history wrapper
class observersation_state_history_wrapper:
    def __init__(self, args, n_agents, max_vehicle_num, max_episode_len, max_history_len):
        self.args = args
        self.max_vehicle_num = max_vehicle_num
        self.max_episode_len = max_episode_len
        self.max_history_len = max_history_len
        self.obs_shape = args.obs_shape_single
        self.n_agents = n_agents
        self.n_threads = args.batch_size_run
        self.curr_t = 0
        self.history = None
        self.agent_id = None
        self.obs_vehicle_id = None
        self.history_out = None
        self.history_episode_out = None

    # Agent profile initialize
    # agent_id: Controllable agents' ID in each threads
    # obs_vehicle_id: Observed vehicles' ID by each agent in each threads
    # history: Recorded vehicles' state
    def agent_obs_profile_init(self, obs):
        obs = np.array(obs)

        n_threads, n_agents, obs_num, obs_dim = obs.shape

        self.agent_id = [[] for _ in range(n_threads)]
        self.history = [{} for _ in range(n_threads)]
        self.obs_vehicle_id = []

        for k in range(n_threads):
            for i in range(n_agents):
                agent_id = int(obs[k, i, 0, 0])

                if agent_id not in self.agent_id[k]:
                    self.agent_id[k].append(agent_id)

                self.history[k][i] = {}

            self.obs_vehicle_id.append([[] for _ in range(self.n_agents)])

        return self.history

    # Pure observation and state wrapper w/o ID, for baselines like IPPO and QMIX, multi-threading
    # input: State: Array (n_threads, 1, state_dim)
    # Obs: Array (n_threads, n_agents, obs_num, obs_dim)
    # Output: array, list of array
    def pure_obs_state_wrapper(self, state, obs):
        n_threads, n_agents, obs_num, obs_dim = np.array(obs).shape
        _, _, state_dim = state.shape
        n_vehicles = int(state_dim // obs_dim)
        new_state = state.reshape(n_threads, 1, n_vehicles, obs_dim)

        new_state = new_state[:, :, :, 1:].reshape((n_threads, -1))
        new_obs = obs[:, :, :, 1:].reshape((n_threads, n_agents, -1))
        return new_state, new_obs

    # Multi thread version
    # Create a list with fixed length to store history
    # If new vehicle is observed, create a new deque
    # If existing vehicle is observed, append to existing deque
    # If vehicle is not observed this time, append a new zero_list
    # Ego state included in the observation
    def obs_history_create(self, obs):
        obs = np.array(obs)
        n_threads, n_agents, obs_num, obs_dim = obs.shape
        for k in range(n_threads):
            for i in range(n_agents):
                agent_id = int(obs[k, i, 0, 0])
                agent_idx = self.agent_id[k].index(agent_id)

                # Observed vehicle's ID in this time step
                observed_id_list = []
                for j in range(obs_num):
                    if np.any(obs[k, i, j, :]):
                        observed_id = int(obs[k, i, j, 0])
                        observed_id_list.append(observed_id)
                        if observed_id not in self.obs_vehicle_id[k][agent_idx]:
                            self.obs_vehicle_id[k][agent_idx].append(observed_id)
                            observed_idx = self.obs_vehicle_id[k][agent_idx].index(observed_id)
                            self.history[k][agent_idx][observed_idx] = deque(maxlen=self.max_episode_len)
                        else:
                            observed_idx = self.obs_vehicle_id[k][agent_idx].index(observed_id)

                        self.history[k][agent_idx][observed_idx].append(obs[k, i, j, 1:].copy())

                # Non-observed, masked with zeros
                for existing_id in self.obs_vehicle_id[k][agent_idx]:
                    if existing_id not in observed_id_list:
                        existing_idx = self.obs_vehicle_id[k][agent_idx].index(existing_id)
                        self.history[k][agent_idx][existing_idx].append(np.zeros_like(obs[k, i, 0, 1:]))
        return self.agent_id, self.obs_vehicle_id, self.history

    # Multi thread version
    # Output for behavior classification and latent generation
    # Use for action execution in the rollout runner
    def obs_history_output(self):
        self.history_out = np.zeros((self.n_threads, self.n_agents, self.max_vehicle_num,
                                     self.max_history_len, self.obs_shape))

        for k in range(self.n_threads):
            for i in range(self.n_agents):
                agent_id = self.agent_id[k][i]
                agent_idx = i

                observed_id_list = self.obs_vehicle_id[k][agent_idx]
                for observed_id in observed_id_list:
                    observed_idx = observed_id_list.index(observed_id)

                    history = self.history[k][agent_idx][observed_idx]
                    for j in range(min(len(history), self.max_history_len)):
                        self.history_out[k, agent_idx, observed_idx, self.max_history_len - 1 - j, :] \
                            = self.history[k][agent_idx][observed_idx][len(history) - j - 1].copy()

        return self.history_out

    # Multi thread version
    # Output for trajectory prediction when selecting actions
    # Use for action execution in the rollout runner
    def obs_single_history_output(self):
        self.single_history_out = np.zeros((self.n_threads, self.n_agents, self.max_vehicle_num, self.obs_shape))

        for k in range(self.n_threads):
            for i in range(self.n_agents):
                agent_id = self.agent_id[k][i]
                agent_idx = i

                observed_id_list = self.obs_vehicle_id[k][agent_idx]
                for observed_id in observed_id_list:
                    observed_idx = observed_id_list.index(observed_id)

                    history = self.history[k][agent_idx][observed_idx]
                    for j in range(min(len(history), self.max_history_len)):
                        self.single_history_out[k, agent_idx, observed_idx, :] \
                            = self.history[k][agent_idx][observed_idx][-1].copy()

        return self.single_history_out

    # Multi thread version
    # Output for episode history segment and latent VAE training
    def obs_history_episode_output(self, mask):
        # mask: [n_threads, actual episode_len, n_agents]
        mask = np.array(mask)

        num_history = int(self.max_episode_len // self.max_history_len)

        self.raw_history_episode_out = np.zeros((self.n_threads, self.n_agents, self.max_vehicle_num,
                                                 self.max_episode_len, self.obs_shape))

        for k in range(self.n_threads):
            for i in range(self.n_agents):
                agent_id = self.agent_id[k][i]
                agent_idx = i

                observed_id_list = self.obs_vehicle_id[k][agent_idx]
                for observed_id in observed_id_list:
                    observed_idx = observed_id_list.index(observed_id)

                    history = self.history[k][agent_idx][observed_idx]

                    for j in range(len(history)):
                        self.raw_history_episode_out[k, agent_idx, observed_idx, self.max_episode_len - 1 - j, :] \
                            = self.history[k][agent_idx][observed_idx][len(history) - j - 1].copy() \
                              * (mask[k, self.max_episode_len - j - 1, i])

        self.history_episode_out = self.raw_history_episode_out.reshape(
            (self.n_threads, self.n_agents, self.max_vehicle_num, num_history, self.max_history_len, self.obs_shape))

        return self.raw_history_episode_out, self.history_episode_out



import torch
import torch.nn as nn

"""RNN modules."""


class RNNLayer(nn.Module):
    def __init__(self, inputs_dim, outputs_dim, recurrent_N, use_orthogonal):
        super(RNNLayer, self).__init__()
        self._recurrent_N = recurrent_N
        self._use_orthogonal = use_orthogonal

        self.rnn = nn.GRU(inputs_dim, outputs_dim, num_layers=self._recurrent_N, batch_first=True) # (batch_size=2, (recurrent_N=1), n_timesteps=60, rnn_feat_size=64)
        for name, param in self.rnn.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0)
            elif 'weight' in name:
                if self._use_orthogonal:
                    nn.init.orthogonal_(param)
                else:
                    nn.init.xavier_uniform_(param)
        self.norm = nn.LayerNorm(outputs_dim)

    def forward(self, x, hxs):
        x, hxs = self.rnn(x, hxs)
        x = x.squeeze(0)
        hxs = hxs.transpose(0, 1)
        # else:
        #     # x is a (T, N, -1) tensor that has been flatten to (T * N, -1)
        #     N = hxs.size(0)
        #     T = int(x.size(0) / N)

        #     # unflatten
        #     x = x.view(T, N, x.size(1))

        #     # Same deal with masks
        #     # masks = masks.view(T, N)

        #     # Let's figure out which steps in the sequence have a zero for any agent
        #     # We will always assume t=0 has a zero in it as that makes the logic cleaner
        #     has_zeros = ((masks[1:] == 0.0)
        #                  .any(dim=-1)
        #                  .nonzero()
        #                  .squeeze()
        #                  .cpu())

        #     # +1 to correct the masks[1:]
        #     if has_zeros.dim() == 0:
        #         # Deal with scalar
        #         has_zeros = [has_zeros.item() + 1]
        #     else:
        #         has_zeros = (has_zeros + 1).numpy().tolist()

        #     # add t=0 and t=T to the list
        #     has_zeros = [0] + has_zeros + [T]

        #     hxs = hxs.transpose(0, 1)

        #     outputs = []
        #     for i in range(len(has_zeros) - 1):
        #         # We can now process steps that don't have any zeros in masks together!
        #         # This is much faster
        #         start_idx = has_zeros[i]
        #         end_idx = has_zeros[i + 1]
        #         temp = (hxs * masks[start_idx].view(1, -1, 1).repeat(self._recurrent_N, 1, 1)).contiguous()
        #         rnn_scores, hxs = self.rnn(x[start_idx:end_idx], temp)
        #         outputs.append(rnn_scores)

        #     # assert len(outputs) == T
        #     # x is a (T, N, -1) tensor
        #     x = torch.cat(outputs, dim=0)

        #     # flatten
        #     x = x.reshape(T * N, -1)
        #     hxs = hxs.transpose(0, 1)

        x = self.norm(x)
        return x, hxs


import datetime
import os
import pprint
import time
import threading
import torch as th
from types import SimpleNamespace as SN
from utils.logging import Logger
from utils.timehelper import time_left, time_str
from os.path import dirname, abspath
from os import makedirs

from learners.ippo_learner import IPPOLearner

from runners.ippo_parallel_runner import ParallelRunner
from runners.ippo_parallel_runner_mpe import ParallelRunner as ParallelRunner_mpe

from controllers.dcntrl_controller import DcntrlMAC
from components.episode_buffer import ReplayBuffer
from components.transforms import OneHot

from nova.behavior_FC_policy import Behavior_policy as behavior_fc_policy
from nova.stable_behavior_policy import Behavior_policy as soft_behavior_policy
from nova.behavior_policy import Behavior_policy as hard_behavior_policy
from nova.prediction_policy import Prediction_policy

import Heterogeneous_Highway_Env as highway_env
import gym
from envs.env_wrappers import SubprocVecEnv
from envs.env_wrappers_mpe import SubprocVecEnv as SubprocVecEnv_mpe
from envs.mpe.MPE_env import MPEEnv
import numpy as np

def run(_run, _config, _log):

    # check args sanity
    _config = args_sanity_check(_config, _log)

    args = SN(**_config)
    args.device = "cuda" if args.use_cuda else "cpu"

    # Create the local results directory
    if args.local_results_path == "":
        args.local_results_path = dirname(dirname(abspath(__file__)))
    makedirs(args.local_results_path, exist_ok=True)

    # setup loggers
    logger = Logger(_log)

    _log.info("Experiment Parameters:")
    experiment_params = pprint.pformat(_config,
                                       indent=4,
                                       width=1)
    _log.info("\n\n" + experiment_params + "\n")

    # configure tensorboard logger
    date_time = datetime.datetime.now().strftime("%m-%d-%H-%M-%S")
    # f"diff={args.env_args['difficulty']}"
    envargs_list = []  # env args we want to appear in name

    algargs_list = [] # [f"act={args.action_selector}"] # alg args we want to appear in name

    namelist = [args.name, args.env, args.label, *envargs_list, *algargs_list, f"seed={args.seed}", date_time]
    namelist = [name.replace("_", "-") for name in namelist if name is not None]
    args.unique_token = "_".join(namelist) 
    if args.use_tensorboard:
        tb_logs_direc = os.path.join(args.local_results_path, "tb_logs")
        tb_exp_direc = os.path.join(tb_logs_direc, "{}").format(args.unique_token)
        logger.setup_tb(tb_exp_direc)

    # sacred is on by default
    logger.setup_sacred(_run)

    # Run and train
    run_sequential(args=args, logger=logger)

    # Clean up after finishing
    print("Exiting Main")

    print("Stopping all threads")
    for t in threading.enumerate():
        if t.name != "MainThread":
            print("Thread {} is alive! Is daemon: {}".format(t.name, t.daemon))
            t.join(timeout=1)
            print("Thread joined")

    print("Exiting script")

    # Making sure framework really exits
    os._exit(os.EX_OK)


def evaluate_sequential(args, runner, learner):
    for _ in range(args.test_nepisode):
        runner.run(test_mode=True)

    if args.save_replay:
        runner.save_replay()

    runner.close_env()

def get_model_path(checkpoint_path, load_step):
    timesteps = []
    timestep_to_load = 0
    # Go through all files in args.checkpoint_path
    for name in os.listdir(checkpoint_path):
        full_name = os.path.join(checkpoint_path, name)
        # Check if they are dirs the names of which are numbers
        if os.path.isdir(full_name) and name.isdigit():
            timesteps.append(int(name))

    if load_step == 0:
        # choose the max timestep
        timestep_to_load = max(timesteps)
    else:
        # choose the timestep closest to load_step
        timestep_to_load = min(timesteps, key=lambda x: abs(x - load_step))
    model_path = os.path.join(checkpoint_path, str(timestep_to_load))
    print("MODEL PATH IS ", model_path)
    return model_path

def run_sequential(args, logger):
    # Initialize the environment wrapper
    if args.env == "highway":
        env = make_train_env(args)
    else:
        env = make_train_env_mpe(args)

    # Initialize the runner
    if args.env == "highway":
        runner = ParallelRunner(args=args, env=env, logger=logger)
    else:
        runner = ParallelRunner_mpe(args=args, env=env, logger=logger)

    # Set up schemes and groups here
    env_info = runner.get_env_info(args)
    args.n_agents = env_info["n_agents"]
    args.n_actions = env_info["n_actions"]
    args.state_shape = env_info["state_shape"]
    args.obs_shape = env_info["obs_shape"]
    args.episode_limit = env_info["episode_limit"]

    # Initialize the number of entities within the environment
    if args.env == "highway":
        args.max_vehicle_num = args.n_other_vehicles + args.n_agents
    else:
        args.max_vehicle_num = args.num_landmarks + args.n_agents + args.num_random_agents

    # Initialize the folder to store the screenshots
    if args.animation_enable:
        if not os.path.isdir("animation"):
            os.makedirs("animation")
        for i in range(args.batch_size_run):
            if not os.path.isdir("animation/" + str(i)):
                os.makedirs("animation/" + str(i))

    print("EPISODE LIMIT IS ", env_info["episode_limit"])
    
    # Default/Base scheme
    scheme = {
        "state": {"vshape": env_info["state_shape"]},
        "obs": {"vshape": env_info["obs_shape"], "group": "agents"},
        "actions": {"vshape": (1,), "group": "agents", "dtype": th.long},
        "rnn_states_actors": {"vshape": (args.rnn_hidden_dim,), "group": "agents"},
        "rnn_states_critics": {"vshape": (args.rnn_hidden_dim,), "group": "agents"},

        # Processed observation for every vehicle
        "history": {"vshape": (args.max_vehicle_num, args.obs_shape_single,), "group": "agents"},
        # Behavioral incentive
        "behavior_latent": {"vshape": (args.max_vehicle_num, args.latent_dim,), "group": "agents"},
        # Instant incentive
        "attention_latent": {"vshape": (args.max_vehicle_num, args.attention_dim,), "group": "agents"},

        "avail_actions": {"vshape": (env_info["n_actions"],), "group": "agents", "dtype": th.int},
        "reward": {"vshape": (1,), "group": "agents"},

        # Vehicle speed, for navigation metric computation
        "speed": {"vshape": (1,), "group": "agents"},
        "terminated": {"vshape": (1,), "group": "agents", "dtype": th.uint8},
    }
    groups = {
        "agents": args.n_agents
    }
    preprocess = {
        "actions": ("actions_onehot", [OneHot(out_dim=args.n_actions)])
    }

    buffer = ReplayBuffer(scheme, groups, args.buffer_size, env_info["episode_limit"] + 1,
                          preprocess=preprocess,
                          device="cpu" if args.buffer_cpu_only else args.device)
    # print("CREATED REPLAY BUFFER")

    # Setup multi-agent controller here
    mac = DcntrlMAC(buffer.scheme, groups, args)

    # Learner
    learner = IPPOLearner(mac, buffer.scheme, logger, args)

    # Initialize behavioral incentive inference
    if args.Behavior_enable:
        if args.behavior_fully_connected:
            # Fully connected behavioral incentive inference module
            behavior_learner = behavior_fc_policy(args, logger)
        elif args.soft_update_enable:
            # Behavioral incentive inference module with soft update policy (iPLAN)
            behavior_learner = soft_behavior_policy(args, logger)
        else:
            # Behavioral incentive inference module with hard update policy
            behavior_learner = hard_behavior_policy(args, logger)
    else:
        behavior_learner = None

    # Initialize instant incentive inference
    if args.GAT_enable:
        # Instant incentive inference module
        prediction_learner = Prediction_policy(args, logger)
    else:
        prediction_learner = None

    # Give runner the scheme
    runner.setup(scheme=scheme, groups=groups, preprocess=preprocess,
                 mac=mac, behavior_learner=behavior_learner, prediction_learner=prediction_learner)

    if args.use_cuda:
        learner.cuda()

    # Load models
    if args.checkpoint_paths[0] != "":
        # iterate thru checkpoint paths and args.load_steps to get the model paths
        model_paths = []
        for checkpoint_path in args.checkpoint_paths:
            if not os.path.isdir(checkpoint_path):
                logger.console_logger.info("Checkpoint directory {} doesn't exist".format(checkpoint_path))
                return
            model_path = get_model_path(checkpoint_path, args.load_step) # loads the nearest step to args.load_step
            model_paths.append(model_path)

        logger.console_logger.info("Loading models from {}".format(model_paths))
        learner.load_models(model_paths)

        if args.Behavior_enable:
            behavior_learner.load_models(model_paths)

        if args.GAT_enable:
            prediction_learner.load_models(model_paths)

        # runner.t_env = timestep_to_load

    # start training
    episode = 0
    last_test_T = -args.test_interval - 1
    last_log_T = 0
    model_save_time = 0
    agent_batch_save_time = 0

    start_time = time.time()
    last_time = start_time

    logger.console_logger.info("Beginning training for {} timesteps".format(args.t_max))

    while runner.t_env <= args.t_max:
        # Run for a whole episode at a time
        episode_batch, avg_win_rates, avg_rwd, avg_len\
            = runner.run(test_mode=False)
        # Insert the episode batch into the replay buffer
        learner.insert_episode_batch(episode_batch)

        # Update the behavior incentive inference learner
        if args.Behavior_enable:
            if runner.t_env > args.Behavior_warmup:
                behavior_loss, stability_loss, total_loss = behavior_learner.learn(episode_batch, runner.t_env)
            else:
                behavior_loss, stability_loss, total_loss = [0], [0], [0]
        else:
            behavior_loss, stability_loss, total_loss = None, None, None

        # Update the instant incentive inference learner
        if args.GAT_enable:
            if runner.t_env > args.GAT_warmup:
                prediction_loss = prediction_learner.learn(episode_batch, runner.t_env)
            else:
                prediction_loss = [0]
        else:
            prediction_loss = None

        learner.train(runner.t_env)

        # Execute test runs once in a while
        if (runner.t_env - last_test_T) / args.test_interval >= 1.0:

            logger.console_logger.info("t_env: {} / {}".format(runner.t_env, args.t_max))
            logger.console_logger.info("Estimated time left: {}. Time passed: {}".format(
                time_left(last_time, last_test_T, runner.t_env, args.t_max), time_str(time.time() - start_time)))
            last_time = time.time()

            last_test_T = runner.t_env

        # save models
        if args.save_model and (runner.t_env - model_save_time >= args.save_model_interval or model_save_time == 0):
            model_save_time = runner.t_env
            save_path = os.path.join(args.local_results_path, "models", args.unique_token, str(runner.t_env))
            os.makedirs(save_path, exist_ok=True)
            logger.console_logger.info("Saving models to {}".format(save_path))

            # learner should handle saving/loading -- delegate actor save/load to,
            # use appropriate filenames to do critics, optimizer states
            learner.save_models(save_path)
            if args.Behavior_enable:
                behavior_learner.save_models(save_path)
            if args.GAT_enable:
                prediction_learner.save_models(save_path)

        if (runner.t_env - last_log_T) >= args.log_interval:
            logger.log_stat("episode", episode, runner.t_env)
            logger.print_recent_stats()
            last_log_T = runner.t_env

        log_print(args, episode, runner.t_env, avg_win_rates, avg_rwd, avg_len, behavior_loss,
                  stability_loss, total_loss, prediction_loss)

        # Enable the navigation metrics printing for highway_env
        if args.metrics_enable and args.env == "highway":
            print("Navigation Metrics Print")
            print("--------------------------------")
            for i in range(int(args.num_test_episodes // args.batch_size_run)):
                # Run for a whole episode at a time
                avg_speed, avg_survival_time, win_num = metric_generate(episode_batch)
                for j in range(args.batch_size_run):
                    episode_test = args.batch_size_run * i + j
                    metric_log_print(args, episode_test, avg_speed[j], avg_survival_time[j], win_num[j])

        episode += args.batch_size_run

    runner.close_env()
    logger.console_logger.info("Finished Training")

# Helper function to generate the metrics of Highway
def metric_generate(episode_batch):
    speed_raw = episode_batch["speed"][:, :-1]
    n_threads, episode_len, n_agents, _ = speed_raw.shape
    speed_raw = speed_raw.reshape(n_threads, episode_len, n_agents)

    terminate_mask_raw = 1 - episode_batch["terminated"][:, :-1]
    terminate_mask_raw = terminate_mask_raw.reshape(n_threads, episode_len, n_agents)

    avg_speed = speed_raw * terminate_mask_raw

    avg_survival_time = terminate_mask_raw.sum(axis=1)

    win_num = np.zeros((n_threads,))
    for i in range(n_threads):
        win_num[i] = sum([avg_survival_time[i, j] == episode_len for j in range(n_agents)])

    avg_speed = avg_speed.sum(axis=1) / (avg_survival_time + 1E-10)
    avg_speed = np.average(np.array(avg_speed.cpu().numpy()), axis=1)

    avg_survival_time = np.average(np.array(avg_survival_time.cpu().numpy()), axis=1)
    return avg_speed, avg_survival_time, win_num

# Print the log for each episode
def log_print(args, episode, t_env, episode_win_rate, episode_reward, episode_len, behavior_loss=None,
              stability_loss=None, total_loss=None, prediction_loss=None):
    # log printer for highway_env
    if args.env == "highway":
        print("Episode #", episode,
              " | Current time step: ", t_env,
              " | Average Episode Win Num: ", episode_win_rate,
              " | Average Episode Reward: ", episode_reward,
              " | Average Episode Length: ", episode_len)
        if args.Behavior_enable:
            print("Behavior Loss", sum(behavior_loss),
                  " | Stability Loss", sum(stability_loss),
                  " | Behavior Total Loss", sum(total_loss))
        if args.GAT_enable:
            print("Prediction Loss", sum(prediction_loss))
        print("--------------------------------")
    # log printer for MPE
    elif args.env == "MPE":
        print("Episode #", episode,
              " | Current time step: ", t_env,
              " | Average Episode Reward: ", episode_reward,
              " | Average Episode Length: ", episode_len)
        if args.Behavior_enable:
            print("Behavior Loss", sum(behavior_loss),
                  " | Stability Loss", sum(stability_loss),
                  " | Behavior Total Loss", sum(total_loss))
        if args.GAT_enable:
            print("Prediction Loss", sum(prediction_loss))
        print("--------------------------------")

# Print the navigation metrics
def metric_log_print(args, episode, avg_speed, avg_survival_time, win_num):
    print("Episode #", episode,
          " | Average Episode Win Num: ", win_num,
          " | Average Episode Survival Time: ", avg_survival_time,
          " | Average Episode Speed: ", avg_speed)
    print("--------------------------------")

def args_sanity_check(config, _log):
    # set CUDA flags
    # config["use_cuda"] = True # Use cuda whenever possible!
    if config["use_cuda"] and not th.cuda.is_available():
        config["use_cuda"] = False
        _log.warning("CUDA flag use_cuda was switched OFF automatically because no CUDA devices are available!")

    return config

# Create the parallel training environment for highway-env
def make_train_env(all_args):
    def get_env_fn(rank):
        def init_env():
            env = env_wrapper(all_args)
            env.seed(all_args.seed + rank * 1000)
            return env
        return init_env

    if all_args.batch_size_run == 1:
        return SubprocVecEnv([get_env_fn(0)])
    else:
        return SubprocVecEnv([get_env_fn(i) for i in range(all_args.batch_size_run)])

# Create the parallel training environment for MPE
def make_train_env_mpe(all_args):
    def get_env_fn(rank):
        def init_env():
            env = MPEEnv(all_args)
            env.seed(all_args.seed + rank * 1000)
            return env
        return init_env

    if all_args.batch_size_run == 1:
        return SubprocVecEnv_mpe([get_env_fn(0)])
    else:
        return SubprocVecEnv_mpe([get_env_fn(i) for i in range(all_args.batch_size_run)])

# env wrapper for highway_env
def env_wrapper(args):
    if args.difficulty == "easy":
        env = gym.make("highway-hetero-v0")
    else:
        env = gym.make("highway-hetero-H-v0")

    env.configure({
      "action": {
        "type": "MultiAgentAction",
        "action_config": {
          "type": "DiscreteMetaAction",
        }
      }
    })

    env.configure({
        "observation": {
            "type": "MultiAgentObservation",
            "observation_config": {
                "type": "Kinematics",
                "vehicles_count": args.n_obs_vehicles,
                "features": ["id", "presence", "x", "y", "vx", "vy"],
                "features_range": {
                    "x": [-100, 100],
                    "y": [-100, 100],
                    "vx": [-20, 20],
                    "vy": [-20, 20]
                },
                "absolute": False,
                "order": "sorted"
            }
        }
    })
    # State definition
    env.configure({"features": ["id", "presence", "x", "y", "vx", "vy"]})

    env.configure({"duration": args.episode_limit})
    env.configure({"lanes_count": args.n_lane})
    env.configure({"controlled_vehicles": args.n_agents})
    env.configure({"simulation_frequency": 5})
    # Display
    env.configure({'scaling': args.scaling, 'screen_height': args.screen_height, 'screen_width': args.screen_width})
    env.configure({"vehicles_count": args.n_other_vehicles})
    env = highway_env.envs.MultiAgentWrapper(env)
    return env


import copy
import numpy as np
import math
import torch
import torch.nn as nn

def init(module, weight_init, bias_init, gain=1):
    weight_init(module.weight.data, gain=gain)
    bias_init(module.bias.data)
    return module

def get_clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for i in range(N)])

def check(input):
    output = torch.from_numpy(input) if type(input) == np.ndarray else input
    return output

def get_grad_norm(it):
    sum_grad = 0
    for x in it:
        if x.grad is None:
            continue
        sum_grad += x.grad.norm() ** 2
    return math.sqrt(sum_grad)

def update_linear_schedule(optimizer, epoch, total_num_epochs, initial_lr):
    """Decreases the learning rate linearly"""
    lr = initial_lr - (initial_lr * (epoch / float(total_num_epochs)))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr

def huber_loss(e, d):
    a = (abs(e) <= d).float()
    b = (e > d).float()
    return a*e**2/2 + b*d*(abs(e)-d/2)

def mse_loss(e):
    return e**2/2

def get_shape_from_obs_space(obs_space):
    if obs_space.__class__.__name__ == 'Box':
        obs_shape = obs_space.shape
    elif obs_space.__class__.__name__ == 'list':
        obs_shape = obs_space
    else:
        raise NotImplementedError
    return obs_shape

def get_shape_from_act_space(act_space):
    if act_space.__class__.__name__ == 'Discrete':
        act_shape = 1
    elif act_space.__class__.__name__ == "MultiDiscrete":
        act_shape = act_space.shape
    elif act_space.__class__.__name__ == "Box":
        act_shape = act_space.shape[0]
    elif act_space.__class__.__name__ == "MultiBinary":
        act_shape = act_space.shape[0]
    else:  # agar
        act_shape = act_space[0].shape[0] + 1  
    return act_shape


import os
import json
import struct
import io
import numpy as np

# to read rmappo files
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator
import tensorboard.compat.proto.event_pb2 as event_pb2  # to read all other files


def newest(path):
    '''Returns full path of newest file in given path'''
    files = os.listdir(path)
    paths = [os.path.join(path, basename) for basename in files]
    return max(paths, key=os.path.getctime)


def read(data):
    header = struct.unpack('Q', data[:8])
    event_str = data[12:12 + int(header[0])]  # 8+4
    data = data[12 + int(header[0]) + 4:]
    return data, event_str


def get_tb_stats(logdir, desired_tag):
    """Gets newest events file from logdir and returns specified statistic and timesteps"""
    try:
        logpath = newest(logdir)
        with open(logpath, 'rb') as f:
                        data = f.read()
    except FileNotFoundError:
        print('Unable to find log file in ', logdir)
        return

    steps, tag_values = [], []
    while data:
        data, event_str = read(data)
        event = event_pb2.Event()

        event.ParseFromString(event_str)
        if event.HasField('summary'):
            for value in event.summary.value:
                if value.HasField('simple_value'):
                    if value.tag == desired_tag:
                        steps.append(event.step)
                        tag_values.append(value.simple_value)
    return {"ts": steps,
            desired_tag: tag_values
            }


def get_mappo_tb_stats(logdir, desired_tag):
    it = EventAccumulator(logdir).Reload()
    events = it.Scalars(desired_tag)
    returns = np.array([e.value for e in events])
    ts = np.array([e.step for e in events])

    return {
            "ts": ts,
            desired_tag: returns
            }

def get_sacred_stats(logdir, desired_tag):
    f = open(os.path.join(logdir, 'info.json'))
    data = json.load(f)
    ts = data["test_ippo_battle_won_mean_T"]
    returns = data["test_ippo_battle_won_mean"]

    return {
            "ts": ts,
            desired_tag: returns
            }

def combine_tb_logs(logdirs,
                    stat_name="test_ippo_battle_won_mean",
                    ts_round_base=None,
                    results_for_sns=True):
    '''
    Reads tensorboard stats from multiple log files and averages them. 
    All log files should have same number of evaluations at approximately the same times

    logdirs: a list of string paths, WITHOUT the events file specified. Newest events file will be read from 
             specified path by default. 
    stat_name: name of statistic to be read from the events file
    ts_round_base: stats are logged after the episode ends, so the statistic logging timestep does not match the log_interval 
                   specifid in the .yaml, but is close. If not none, this function will round the ts to the nearest ts_round_base. 
    '''
    all_results_dict = {}
    ts, max_ts = [], 0
    for i, logdir in enumerate(logdirs):
        if "rmappo" in logdir:
            res_dict = get_mappo_tb_stats(logdir, desired_tag=stat_name)
        elif "sacred" in logdir:
            res_dict = get_sacred_stats(logdir, desired_tag=stat_name)
        else:
            res_dict = get_tb_stats(logdir, desired_tag=stat_name)
        all_results_dict[f"run_{i}"] = res_dict[stat_name]

        ts_list = res_dict["ts"]
        if len(ts_list) >= max_ts:
            ts = np.array(ts_list)
            max_ts = len(ts_list)

    if ts_round_base is not None:
        # round a_number to the nearest base
        def nearest_multiple(a_number, base): return base * \
            round(a_number / base)
        ts = np.array([nearest_multiple(step, ts_round_base) for step in ts])
    all_results_dict["ts"] = ts

    if not results_for_sns:
        results_all = np.array(
            [all_results_dict[f"run_{i}"] for i in range(len(logdirs))])
        mean_results, std_results = np.mean(
            results_all, axis=0), np.std(results_all, axis=0)
        return (all_results_dict["ts"], mean_results, std_results)

    return all_results_dict



import numpy as np


class DecayThenFlatSchedule():

    def __init__(self,
                 start,
                 finish,
                 time_length,
                 decay="exp"):

        self.start = start
        self.finish = finish
        self.time_length = time_length
        self.delta = (self.start - self.finish) / self.time_length
        self.decay = decay

        if self.decay in ["exp"]:
            assert self.start >= self.finish, "Only exponential decay is currently supported."
            self.exp_scaling = (-1) * self.time_length / np.log(self.finish) if self.finish > 0 else 1

    def eval(self, T):
        if self.decay in ["linear"]:
            if self.start < self.finish:
                return min(self.finish, self.start - self.delta * T)
            else:
                return max(self.finish, self.start - self.delta * T)
        elif self.decay in ["exp"]:
            return min(self.start, max(self.finish, np.exp(- T / self.exp_scaling)))
    pass


class FlatSchedule():

    def __init__(self,
                 epsilon,
                 noise_coef=None,
                 noise_decay="linear",
                 time_length=None):

        self.epsilon = epsilon
        self.noise_coef = noise_coef 
        self.noise_decay = noise_decay # decay the noise coef to 0 linearly  
        self.time_length = time_length

        if self.noise_coef is not None:
            self.noise_delta = self.noise_coef / self.time_length

    def eval(self, T):
        if self.noise_coef is not None: 
            if self.noise_decay in ["linear"]:
                noise = np.random.rand() * max((self.noise_coef - self.noise_delta * T), 0)
            return self.epsilon - noise

        return self.epsilon
    pass




import math
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

class PopArt(torch.nn.Module):
    
    def __init__(self, input_shape, output_shape, norm_axes=1, beta=0.99999, epsilon=1e-5, device=torch.device("cpu")):
        
        super(PopArt, self).__init__()

        self.beta = beta
        self.epsilon = epsilon
        self.norm_axes = norm_axes
        self.tpdv = dict(dtype=torch.float32, device=device)

        self.input_shape = input_shape
        self.output_shape = output_shape

        self.weight = nn.Parameter(torch.Tensor(output_shape, input_shape)).to(**self.tpdv)
        self.bias = nn.Parameter(torch.Tensor(output_shape)).to(**self.tpdv)
        
        self.stddev = nn.Parameter(torch.ones(output_shape), requires_grad=False).to(**self.tpdv)
        self.mean = nn.Parameter(torch.zeros(output_shape), requires_grad=False).to(**self.tpdv)
        self.mean_sq = nn.Parameter(torch.zeros(output_shape), requires_grad=False).to(**self.tpdv)
        self.debiasing_term = nn.Parameter(torch.tensor(0.0), requires_grad=False).to(**self.tpdv)

        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in)
            torch.nn.init.uniform_(self.bias, -bound, bound)
        self.mean.zero_()
        self.mean_sq.zero_()
        self.debiasing_term.zero_()

    def forward(self, input_vector):
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)

        return F.linear(input_vector, self.weight, self.bias)
    
    @torch.no_grad()
    def update(self, input_vector):
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)
        
        old_mean, old_stddev = self.mean, self.stddev

        batch_mean = input_vector.mean(dim=tuple(range(self.norm_axes)))
        batch_sq_mean = (input_vector ** 2).mean(dim=tuple(range(self.norm_axes)))

        self.mean.mul_(self.beta).add_(batch_mean * (1.0 - self.beta))
        self.mean_sq.mul_(self.beta).add_(batch_sq_mean * (1.0 - self.beta))
        self.debiasing_term.mul_(self.beta).add_(1.0 * (1.0 - self.beta))

        self.stddev = (self.mean_sq - self.mean ** 2).sqrt().clamp(min=1e-4)

        self.weight = self.weight * old_stddev / self.stddev
        self.bias = (old_stddev * self.bias + old_mean - self.mean) / self.stddev

    def debiased_mean_var(self):
        debiased_mean = self.mean / self.debiasing_term.clamp(min=self.epsilon)
        debiased_mean_sq = self.mean_sq / self.debiasing_term.clamp(min=self.epsilon)
        debiased_var = (debiased_mean_sq - debiased_mean ** 2).clamp(min=1e-2)
        return debiased_mean, debiased_var

    def normalize(self, input_vector):
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)

        mean, var = self.debiased_mean_var()
        out = (input_vector - mean[(None,) * self.norm_axes]) / torch.sqrt(var)[(None,) * self.norm_axes]
        
        return out

    def denormalize(self, input_vector):
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)

        mean, var = self.debiased_mean_var()
        out = input_vector * torch.sqrt(var)[(None,) * self.norm_axes] + mean[(None,) * self.norm_axes]
        
        out = out.cpu().numpy()

        return out


import torch as th


def build_td_lambda_targets(rewards, terminated, mask, target_qs, n_agents, gamma, td_lambda):
    # Assumes  <target_qs > in B*T*A and <reward >, <terminated >, <mask > in (at least) B*T-1*1
    # Initialise  last  lambda -return  for  not  terminated  episodes
    ret = target_qs.new_zeros(*target_qs.shape)
    ret[:, -1] = target_qs[:, -1] * (1 - th.sum(terminated, dim=1))
    # Backwards  recursive  update  of the "forward  view"
    for t in range(ret.shape[1] - 2, -1,  -1):
        ret[:, t] = td_lambda * gamma * ret[:, t + 1] + mask[:, t] \
                    * (rewards[:, t] + (1 - td_lambda) * gamma * target_qs[:, t + 1] * (1 - terminated[:, t]))
    # Returns lambda-return from t=0 to t=T-1, i.e. in B*T-1*A
    return ret[:, 0:-1]




import time
import numpy as np


def print_time(start_time, T, t_max, episode, episode_rewards):
    time_elapsed = time.time() - start_time
    T = max(1, T)
    time_left = time_elapsed * (t_max - T) / T
    # Just in case its over 100 days
    time_left = min(time_left, 60 * 60 * 24 * 100)
    last_reward = "N\A"
    if len(episode_rewards) > 5:
        last_reward = "{:.2f}".format(np.mean(episode_rewards[-50:]))
    print("\033[F\033[F\x1b[KEp: {:,}, T: {:,}/{:,}, Reward: {}, \n\x1b[KElapsed: {}, Left: {}\n".format(episode, T, t_max, last_reward, time_str(time_elapsed), time_str(time_left)), " " * 10, end="\r")


def time_left(start_time, t_start, t_current, t_max):
    if t_current >= t_max:
        return "-"
    time_elapsed = time.time() - start_time
    t_current = max(1, t_current)
    time_left = time_elapsed * (t_max - t_current) / (t_current - t_start)
    # Just in case its over 100 days
    time_left = min(time_left, 60 * 60 * 24 * 100)
    return time_str(time_left)


def time_str(s):
    """
    Convert seconds to a nicer string showing days, hours, minutes and seconds
    """
    days, remainder = divmod(s, 60 * 60 * 24)
    hours, remainder = divmod(remainder, 60 * 60)
    minutes, seconds = divmod(remainder, 60)
    string = ""
    if days > 0:
        string += "{:d} days, ".format(int(days))
    if hours > 0:
        string += "{:d} hours, ".format(int(hours))
    if minutes > 0:
        string += "{:d} minutes, ".format(int(minutes))
    string += "{:d} seconds".format(int(seconds))
    return string



import torch as th


class Transform:
    def transform(self, tensor):
        raise NotImplementedError

    def infer_output_info(self, vshape_in, dtype_in):
        raise NotImplementedError


class OneHot(Transform):
    def __init__(self, out_dim):
        self.out_dim = out_dim

    def transform(self, tensor):
        y_onehot = tensor.new(*tensor.shape[:-1], self.out_dim).zero_()
        y_onehot.scatter_(-1, tensor.long(), 1)
        return y_onehot.float()

    def infer_output_info(self, vshape_in, dtype_in):
        return (self.out_dim,), th.float32


import torch as th
import numpy as np
from collections import deque


def _flatten(T, N, x):
    return x.reshape(T * N, *x.shape[2:])


def _cast(x):
    return x.transpose(1,0,2).reshape(-1, *x.shape[2:])


class SeparatedReplayBuffer(object):
    def __init__(self, args, obs_shape, share_obs_shape 
                 ):
        self.episode_limit = args.episode_limit   # max ep len
        self.buffer_size = args.buffer_size  # originally n_rollout_threads
        self.thread_id = 0
        self.rnn_hidden_dim = args.rnn_hidden_dim
        self.recurrent_N = args.recurrent_N
        self.batch_size_run = args.batch_size_run

        self.state = deque(maxlen=self.buffer_size)
        self.obs = deque(maxlen=self.buffer_size)
        self.rnn_states_actor = deque(maxlen=self.buffer_size)
        self.rnn_states_critic = deque(maxlen=self.buffer_size)
        self.actions = deque(maxlen=self.buffer_size)
        self.actions_onehot = deque(maxlen=self.buffer_size)
        self.reward = deque(maxlen=self.buffer_size)
        self.terminated_masks = deque(maxlen=self.buffer_size)
        self.active_masks = deque(maxlen=self.buffer_size)
        self.available_actions = deque(maxlen=self.buffer_size)

        self.histories = deque(maxlen=self.buffer_size)
        self.behavior_latent = deque(maxlen=self.buffer_size)

        self.attention_latent = deque(maxlen=self.buffer_size)

    def can_sample(self):
        # print("LEN SELF OBS IS ", len(self.obs))
        # print("LEN BUFFER SIZE IS ", self.buffer_size)
        return len(self.obs) == self.buffer_size

    def data_wrapper(self, element, idx):
        return element[idx].reshape(1, *element.shape[1:])

    def insert(self, state, obs, rnn_states_actor, rnn_states_critic, 
               actions, actions_onehot, rewards, terminated_masks,
               histories, behavior_latent, attention_latent, active_masks=None, available_actions=None):

        for i in range(self.batch_size_run):
            self.state.append(self.data_wrapper(state, i))
            self.obs.append(self.data_wrapper(obs, i))
            self.rnn_states_actor.append(self.data_wrapper(rnn_states_actor, i))
            self.rnn_states_critic.append(self.data_wrapper(rnn_states_critic, i))
            self.actions.append(self.data_wrapper(actions, i))
            self.actions_onehot.append(self.data_wrapper(actions_onehot, i))
            self.reward.append(self.data_wrapper(rewards, i))
            self.terminated_masks.append(self.data_wrapper(terminated_masks, i))

            self.histories.append(self.data_wrapper(histories, i))
            self.behavior_latent.append(self.data_wrapper(behavior_latent, i))

            self.attention_latent.append(self.data_wrapper(attention_latent, i))

            if active_masks is not None:
                self.active_masks.append(self.data_wrapper(active_masks, i))
            if available_actions is not None:
                self.available_actions.append(self.data_wrapper(available_actions, i))

    def get_batch(self):
        if len(self.obs) < self.buffer_size:
            return
        batch = {}
        batch['obs'] = th.cat(list(self.obs))
        batch['state'] = th.cat(list(self.state))
        batch['actions'] = th.cat(list(self.actions))
        batch['actions_onehot'] = th.cat(list(self.actions_onehot))
        batch['rnn_states_actor'] = th.cat(list(self.rnn_states_actor))
        batch['rnn_states_critic'] = th.cat(list(self.rnn_states_critic))
        batch['reward'] = th.cat(list(self.reward))
        batch['terminated_masks'] = th.cat(list(self.terminated_masks))

        batch['history'] = th.cat(list(self.histories))
        batch['behavior_latent'] = th.cat(list(self.behavior_latent))

        batch['attention_latent'] = th.cat(list(self.attention_latent))

        if len(self.active_masks) != 0:
            batch['active_masks'] = th.cat(list(self.active_masks))
        if len(self.available_actions) != 0:
            batch['available_actions'] = th.cat(list(self.available_actions))

        return batch

    def clear_buffer(self):
        self.state.clear()
        self.obs.clear()
        self.rnn_states_actor.clear()
        self.rnn_states_critic.clear()
        self.actions.clear()
        self.actions_onehot.clear()
        self.reward.clear()
        self.terminated_masks.clear()
        self.available_actions.clear()

        self.histories.clear()
        self.behavior_latent.clear()

        self.attention_latent.clear()
